from __future__ import annotations

from typing import TYPE_CHECKING

import asyncio
import grpc.aio

from ..core.logger import core_logger, notification_logger

if TYPE_CHECKING:
	from typing import Optional, Set, AsyncIterator, Dict
	from ..internal import types
	from .Notifications import NOTIFICATIONS
	from .GenericNotificationListener import GenericNotificationListener
	from ..core.OlvidClient import OlvidClient


class ClientListenerHolder:
	def __init__(self, client: "OlvidClient"):
		self._client: "OlvidClient" = client

		self._registered_listeners: Dict[str, Set[GenericNotificationListener]] = {}
		self._iterators_tasks: Dict[str, asyncio.Task] = {}
		self._listener_removed_event: asyncio.Event = asyncio.Event()

		self._stream_listener_task: Optional[asyncio.Task] = None

	#####
	# Public api
	#####
	async def await_task(self):
		if self._stream_listener_task is not None:
			await self._stream_listener_task

	async def wait_for_listener_removed_event(self):
		await self._listener_removed_event.wait()

	async def stop(self):
		# cancel every background task
		for notif in self._iterators_tasks.keys():
			self._iterators_tasks.get(notif).cancel()
		# cancel listening task
		if self._stream_listener_task:
			self._stream_listener_task.cancel()
		# wait for listening task end
		if self._stream_listener_task:
			await self._stream_listener_task

	def add_listener(self, listener: GenericNotificationListener):
		# check listener count is valid (if count == 0 on start it will stay active forever)
		if listener.count == 0:
			print(f"adding listener: {listener}")
			return

		# add listener to registered_listeners
		if not self._registered_listeners.get(listener.listener_key):
			self._registered_listeners[listener.listener_key] = set()
		self._registered_listeners[listener.listener_key].add(listener)

		# create an iterator if needed
		if not self._iterators_tasks.get(listener.listener_key):
			self._add_iterator(listener)

	def remove_listener(self, listener: GenericNotificationListener):
		# remove listener from registered_listeners
		if self._registered_listeners.get(listener.listener_key) \
					and listener in self._registered_listeners.get(listener.listener_key):
			self._registered_listeners[listener.listener_key].remove(listener)
			# mark listener as finished
			listener.mark_as_finished()
			# notify any waiting client that a listener finished, they will check if their listeners list was updated
			# we set and clear event because we only need to wake up methods waiting for this event, and this event might be used multiple times
			self._listener_removed_event.set()
			self._listener_removed_event.clear()

		# remove iterator if it's not used anymore
		if not self._registered_listeners.get(listener.listener_key):
			self._remove_iterator(listener.listener_key)

	def get_listeners_count(self) -> int:
		count = 0
		for listeners_list in self._registered_listeners.values():
			count += len(listeners_list)
		return count

	#####
	# notifications iterators aggregation
	#####
	def _add_iterator(self, listener: GenericNotificationListener):
		# if iterator already exists skip
		if self._iterators_tasks.get(listener.listener_key):
			return
		# create a notification channel for given notification type
		# noinspection PyProtectedMember
		notification_iterator = listener._create_iterator(self._client)

		# create a task listening to the notification iterator
		task = asyncio.get_event_loop().create_task(self._iterator_wrapper(listener.listener_key, listener.notification_type, notification_iterator))
		self._iterators_tasks[listener.listener_key] = task

	def _remove_iterator(self, listener_key: str):
		task = self._iterators_tasks.get(listener_key)

		# no iterator to remove
		if not task:
			return

		# cancel current iterator and remove it from tasks
		task.cancel()
		self._iterators_tasks.pop(listener_key)

		# if there are no more iterators stop listening task
		if len(self._iterators_tasks) == 0 and self._stream_listener_task:
			self._stream_listener_task.cancel()

	async def _iterator_wrapper(self, listener_key: str, notification_type: NOTIFICATIONS, async_iterator: AsyncIterator):
		try:
			async for notification_message in async_iterator:
				self._dispatch_notification(listener_key, notification_message)
		except asyncio.CancelledError:
			raise
		except grpc.aio.AioRpcError as e:
			if e.code() == e.code().UNAVAILABLE:
				if e.details() == "Socket closed":
					core_logger.error(f"{self.__class__.__name__}: {notification_type}: connection lost")
				else:
					core_logger.error(f"{self.__class__.__name__}: {notification_type}: server error: {e.details()}")
			else:
				core_logger.error(f"{self.__class__.__name__}: {notification_type}: grpc error: {e.code()} {e.details()}")
		except Exception as e:
			core_logger.exception(f"{self.__class__.__name__}: {notification_type}: unexpected exception: {e}")

		# unregister notifications listener ...
		listeners = self._registered_listeners.get(listener_key).copy()
		for listener in listeners:
			self.remove_listener(listener)

	def _dispatch_notification(self, listener_key: str, notification_message: types.NotificationMessageType):
		if not self._registered_listeners.get(listener_key):
			return
		registered_listeners = list(self._registered_listeners.get(listener_key))
		notification_logger.info(f"{self._client.__class__.__name__}: notification received: {listener_key}")
		notification_logger.debug(f"{self._client.__class__.__name__}: notification content:  {listener_key}: {notification_message}")
		# sort by priority higher priority first
		registered_listeners.sort(key=lambda listen: listen.priority, reverse=True)
		for listener in registered_listeners:
			try:
				# clone original notification else a listener can modify notification content for next listeners ...
				# noinspection PyProtectedMember
				self._client.add_background_task(listener.handle_notification(notification_message._clone(), self.remove_listener))
			except Exception as e:
				core_logger.exception(f"ClientListenerHolder: cannot create listener task (this is not supposed to happen !!): {e}")



import asyncio
import secrets
from typing import Coroutine, Callable, Optional, Union, Any

# for compatibility with python 3.10
from typing import TypeVar
# noinspection PyTypeHints
Self = TypeVar("ChatBot")


from .logger import tools_logger

from .. import datatypes
from ..core.OlvidClient import OlvidClient
from ..core.OlvidBot import OlvidBot
from ..listeners.Command import Command
from ..listeners.ListenersImplementation import MessageReceivedListener, DiscussionNewListener


####
# This class helps to implements a basic chat bot that can send a welcome message when it enters a new discussion.
# It also add an help command and send help every time a user send a message in a one to one discussion.
# All these behavior can be modified after creation using help_... and welcome_... methods.
####
class ChatBot(OlvidBot):
	WELCOME_MESSAGE_FACTORY: type = Callable[[datatypes.Discussion], Union[Optional[str], Coroutine[Any, Any, Optional[str]]]]
	HELP_MESSAGE_FACTORY: type = Callable[[int], Union[Optional[str], Coroutine[Any, Any, Optional[str]]]]

	def __init__(self, parent_client: OlvidClient = None):
		super().__init__(parent_client=parent_client)
		# welcome message
		self._welcome_message: Optional[str] = None
		self._welcome_message_factory: Optional[ChatBot.WELCOME_MESSAGE_FACTORY] = None
		self._welcome_send_help: bool = True
		self._welcome_send_in_one_to_one: bool = True
		self._welcome_send_in_groups: bool = True

		# help message
		self._help_message: Optional[str] = None
		self._help_message_factory: Optional[ChatBot.HELP_MESSAGE_FACTORY] = None
		self._help_always_send_in_to_one: bool = True
		self._help_always_send_in_group: bool = False
		self._help_command_prefix: Optional[str] = None

		# help command
		self._help_command: Optional[Command] = Command(regexp_filter="!?help", handler=self._help_command_handler)
		self.add_command(self._help_command)
		self._on_message_listener: MessageReceivedListener = MessageReceivedListener(handler=self._message_received_handler)
		self.add_listener(self._on_message_listener)

		# various
		self.add_listener(DiscussionNewListener(handler=self._discussion_new_handler))

	#####
	# Help configuration
	#####
	def help_set_message(self, help_message: str) -> Self:
		self._help_message = help_message
		self._help_message_factory = None
		return self

	def help_set_message_factory(self, factory: HELP_MESSAGE_FACTORY) -> Self:
		self._help_message = None
		self._help_message_factory = factory
		return self

	def help_disable_command(self) -> Self:
		self.remove_command(command=self._help_command)
		self._help_command = None
		return self

	def help_send_on_invalid_command(self, command_prefix: str = "!") -> Self:
		self._help_command_prefix = command_prefix
		return self

	def help_always_send_in_groups(self) -> Self:
		self._help_always_send_in_group = True
		return self

	def help_do_not_always_send_in_one_to_one(self) -> Self:
		self._help_always_send_in_to_one = False
		return self

	def help_add_custom_checker(self, checker: MessageReceivedListener.CheckerType) -> Self:
		self._on_message_listener.add_checker(checker)
		self._help_command.add_checker(checker)
		return self

	async def _message_received_handler(self, message: datatypes.Message):
		# if message is a valid command do nothing
		if self.tool_is_message_a_valid_command(message):
			return

		# check if we send help on invalid command, and if this was supposed to be a command
		if self._help_command_prefix and message.body.startswith(self._help_command_prefix):
			await self.send_help_message(message.discussion_id)

		discussion: datatypes.Discussion = await self.discussion_get(discussion_id=message.discussion_id)
		if discussion.contact_id and self._help_always_send_in_to_one:
			await self.send_help_message(message.discussion_id)
		elif discussion.group_id and self._help_always_send_in_group:
			await self.send_help_message(message.discussion_id)

	#####
	# Welcome configuration
	#####
	def welcome_set_message(self, welcome_message: str) -> Self:
		self._welcome_message = welcome_message
		self._welcome_message_factory = None
		return self

	def welcome_set_message_factory(self, factory: WELCOME_MESSAGE_FACTORY) -> Self:
		self._welcome_message = None
		self._welcome_message_factory = factory
		return self

	def welcome_do_not_send_in_one_to_one(self) -> Self:
		self._welcome_send_in_one_to_one = False
		return self

	def welcome_do_not_send_in_groups(self) -> Self:
		self._welcome_send_in_groups = False
		return self

	def welcome_do_not_send_help(self) -> Self:
		self._welcome_send_help = False
		return self

	#####
	# Help section
	#####
	# This can be overwritten to build more complex help message
	async def send_help_message(self, discussion_id: int):
		if self._help_message:
			await self.message_send(discussion_id=discussion_id, body=self._help_message)
		elif self._help_message_factory:
			promise = self._help_message_factory(discussion_id)
			if asyncio.iscoroutine(promise):
				help_message: str = await promise
			else:
				help_message: str = promise
			if help_message:
				await self.message_send(discussion_id=discussion_id, body=help_message)

	async def _help_command_handler(self, message: datatypes.Message):
		await self.send_help_message(message.discussion_id)

	#####
	# Welcome section
	#####
	# This can be overwritten to build more complex welcome message
	async def send_welcome_message(self, discussion: datatypes.Discussion):
		if self._welcome_message:
			await discussion.post_message(body=self._welcome_message)
		elif self._welcome_message_factory:
			promise = self._welcome_message_factory(discussion)
			if asyncio.iscoroutine(promise):
				welcome_message: str = await promise
			else:
				welcome_message: str = promise
			if welcome_message:
				await discussion.post_message(body=welcome_message)
			if self._welcome_send_help:
				await self.send_help_message(discussion.id)

	# On discussion creation, create a webhook and post it in discussion
	async def _discussion_new_handler(self, discussion: datatypes.Discussion):
		# handle new groups (we do not need to wait to post in a group)
		if discussion.group_id != 0:
			await self.send_welcome_message(discussion)
			return

		# handle the new contact case
		if discussion.contact_id == 0:
			return

		# check if contact have channels to post a message
		contact = await self.contact_get(contact_id=discussion.contact_id)
		if contact.established_channel_count != 0:
			await self.send_welcome_message(discussion)
			return

		# if contact has no channel a background task that will try to post the welcome message later
		self.add_background_task(self._post_welcome_message_task(discussion), name=f"welcome_bot_{contact.id}")
		tools_logger.info(f"WelcomeBot: waiting for channel establishment, contact_id: {contact.id}")

	# if there are no channel create a task that will wait for up to 10 seconds to post the message
	async def _post_welcome_message_task(self, discussion: datatypes.Discussion):
		count = 0
		max_count = 1000
		while count < max_count:
			await asyncio.sleep(1)
			c = await self.contact_get(contact_id=discussion.contact_id)
			if c.established_channel_count > 0:
				await self.send_welcome_message(discussion)
				return
			tools_logger.info(f"WelcomeBot: still waiting for contact channel: {count + 1}/{max_count}")
			count += 1

	####
	#  tools
	####
	# send a randomized greeting in a discussion
	async def tool_basic_welcome_factory(self, discussion: datatypes.Discussion):
		name: str
		if discussion.contact_id:
			contact: datatypes.Contact = await self.contact_get(contact_id=discussion.contact_id)
			name = contact.details.first_name if contact.details.first_name else contact.display_name
		else:
			group: datatypes.Group = await self.group_get(group_id=discussion.group_id)
			if group.members == 1:
				contact: datatypes.Contact = await self.contact_get(contact_id=group.members[0].contact_id)
				name = contact.details.first_name if contact.details.first_name else contact.display_name
			else:
				name = "everyone"
		greetings: str = secrets.choice(["Hi", "Hello", "Hey"])
		return f"""
{greetings} {name} 👋
		""".strip()

	@staticmethod
	async def tool_ignore_location_message_checker(message: datatypes.Message):
		return message.message_location.type == datatypes.MessageLocation.LocationType.LOCATION_TYPE_UNSPECIFIED

	def tool_is_message_a_valid_command(self, message: datatypes.Message):
		return any([isinstance(listener, Command) and listener.match(message) for listener in self._listeners_set])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: olvid/daemon/command/v1/identity_commands.proto
# Protobuf Python Version: 5.27.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    3,
    '',
    'olvid/daemon/command/v1/identity_commands.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .....olvid.daemon.datatypes.v1 import identity_pb2 as olvid_dot_daemon_dot_datatypes_dot_v1_dot_identity__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/olvid/daemon/command/v1/identity_commands.proto\x12\x17olvid.daemon.command.v1\x1a(olvid/daemon/datatypes/v1/identity.proto\"\x14\n\x12IdentityGetRequest\"V\n\x13IdentityGetResponse\x12?\n\x08identity\x18\x01 \x01(\x0b\x32#.olvid.daemon.datatypes.v1.IdentityR\x08identity\"k\n\x1cIdentityUpdateDetailsRequest\x12K\n\x0bnew_details\x18\x02 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.IdentityDetailsR\nnewDetails\"\x1f\n\x1dIdentityUpdateDetailsResponse\"\x1c\n\x1aIdentityRemovePhotoRequest\"\x1d\n\x1bIdentityRemovePhotoResponse\"Z\n\x1fIdentitySetPhotoRequestMetadata\x12\x1a\n\x08\x66ilename\x18\x01 \x01(\tR\x08\x66ilename\x12\x1b\n\tfile_size\x18\x02 \x01(\x04R\x08\x66ileSize\"\x98\x01\n\x17IdentitySetPhotoRequest\x12V\n\x08metadata\x18\x01 \x01(\x0b\x32\x38.olvid.daemon.command.v1.IdentitySetPhotoRequestMetadataH\x00R\x08metadata\x12\x1a\n\x07payload\x18\x02 \x01(\x0cH\x00R\x07payloadB\t\n\x07request\"\x1a\n\x18IdentitySetPhotoResponse\"L\n\x1bIdentityKeycloakBindRequest\x12-\n\x12\x63onfiguration_link\x18\x01 \x01(\tR\x11\x63onfigurationLink\"\x1e\n\x1cIdentityKeycloakBindResponse\"\x1f\n\x1dIdentityKeycloakUnbindRequest\" \n\x1eIdentityKeycloakUnbindResponse\"3\n\x18IdentitySetApiKeyRequest\x12\x17\n\x07\x61pi_key\x18\x01 \x01(\tR\x06\x61piKey\"`\n\x19IdentitySetApiKeyResponse\x12\x43\n\x07\x61pi_key\x18\x01 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.Identity.ApiKeyR\x06\x61piKey\"T\n#IdentitySetConfigurationLinkRequest\x12-\n\x12\x63onfiguration_link\x18\x01 \x01(\tR\x11\x63onfigurationLink\"k\n$IdentitySetConfigurationLinkResponse\x12\x43\n\x07\x61pi_key\x18\x01 \x01(\x0b\x32*.olvid.daemon.datatypes.v1.Identity.ApiKeyR\x06\x61piKeyB\xc4\x01\n\x1b\x63om.olvid.daemon.command.v1B\x15IdentityCommandsProtoP\x01Z\x0folvid.io/daemon\xa2\x02\x03ODC\xaa\x02\x17Olvid.Daemon.Command.V1\xca\x02\x17Olvid\\Daemon\\Command\\V1\xe2\x02#Olvid\\Daemon\\Command\\V1\\GPBMetadata\xea\x02\x1aOlvid::Daemon::Command::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'olvid.daemon.command.v1.identity_commands_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\033com.olvid.daemon.command.v1B\025IdentityCommandsProtoP\001Z\017olvid.io/daemon\242\002\003ODC\252\002\027Olvid.Daemon.Command.V1\312\002\027Olvid\\Daemon\\Command\\V1\342\002#Olvid\\Daemon\\Command\\V1\\GPBMetadata\352\002\032Olvid::Daemon::Command::V1'
  _globals['_IDENTITYGETREQUEST']._serialized_start=118
  _globals['_IDENTITYGETREQUEST']._serialized_end=138
  _globals['_IDENTITYGETRESPONSE']._serialized_start=140
  _globals['_IDENTITYGETRESPONSE']._serialized_end=226
  _globals['_IDENTITYUPDATEDETAILSREQUEST']._serialized_start=228
  _globals['_IDENTITYUPDATEDETAILSREQUEST']._serialized_end=335
  _globals['_IDENTITYUPDATEDETAILSRESPONSE']._serialized_start=337
  _globals['_IDENTITYUPDATEDETAILSRESPONSE']._serialized_end=368
  _globals['_IDENTITYREMOVEPHOTOREQUEST']._serialized_start=370
  _globals['_IDENTITYREMOVEPHOTOREQUEST']._serialized_end=398
  _globals['_IDENTITYREMOVEPHOTORESPONSE']._serialized_start=400
  _globals['_IDENTITYREMOVEPHOTORESPONSE']._serialized_end=429
  _globals['_IDENTITYSETPHOTOREQUESTMETADATA']._serialized_start=431
  _globals['_IDENTITYSETPHOTOREQUESTMETADATA']._serialized_end=521
  _globals['_IDENTITYSETPHOTOREQUEST']._serialized_start=524
  _globals['_IDENTITYSETPHOTOREQUEST']._serialized_end=676
  _globals['_IDENTITYSETPHOTORESPONSE']._serialized_start=678
  _globals['_IDENTITYSETPHOTORESPONSE']._serialized_end=704
  _globals['_IDENTITYKEYCLOAKBINDREQUEST']._serialized_start=706
  _globals['_IDENTITYKEYCLOAKBINDREQUEST']._serialized_end=782
  _globals['_IDENTITYKEYCLOAKBINDRESPONSE']._serialized_start=784
  _globals['_IDENTITYKEYCLOAKBINDRESPONSE']._serialized_end=814
  _globals['_IDENTITYKEYCLOAKUNBINDREQUEST']._serialized_start=816
  _globals['_IDENTITYKEYCLOAKUNBINDREQUEST']._serialized_end=847
  _globals['_IDENTITYKEYCLOAKUNBINDRESPONSE']._serialized_start=849
  _globals['_IDENTITYKEYCLOAKUNBINDRESPONSE']._serialized_end=881
  _globals['_IDENTITYSETAPIKEYREQUEST']._serialized_start=883
  _globals['_IDENTITYSETAPIKEYREQUEST']._serialized_end=934
  _globals['_IDENTITYSETAPIKEYRESPONSE']._serialized_start=936
  _globals['_IDENTITYSETAPIKEYRESPONSE']._serialized_end=1032
  _globals['_IDENTITYSETCONFIGURATIONLINKREQUEST']._serialized_start=1034
  _globals['_IDENTITYSETCONFIGURATIONLINKREQUEST']._serialized_end=1118
  _globals['_IDENTITYSETCONFIGURATIONLINKRESPONSE']._serialized_start=1120
  _globals['_IDENTITYSETCONFIGURATIONLINKRESPONSE']._serialized_end=1227
# @@protoc_insertion_point(module_scope)
